#include <sivkernel/interface.h>
#include <sivkernel/method.h>
#include <sivkernel/kernel.h>
#include <sivkernel/generated/library.h>
#include <sivdk/exception_kind.h>
#include <sivkernel-mocks/domain_mock.h>
#include <sivdk/interface_handle.h>
#include <gtest/gtest.h>
#include <gmock/gmock.h>
#include <string>
#include <unordered_set>

using namespace sivelkiria::kernel;
using namespace sivelkiria::kernel::generated;
using namespace mocks;
using namespace sivelkiria::common;
using namespace data_pack;
using namespace std;
using namespace testing;


TEST(Invocation, SimpleCall)
{
    auto *const library = library::get_instance();

    kernel kernel(library);
    auto domain1 = make_shared<local_domain_mock>(&kernel);

    const auto *const module_feature = library->get_interface(interface_handle::system__module_feature);
    const method_base *stop_extra_threads = nullptr;
    const method_id stop_extra_threads_id = 2;
    ASSERT_EQ(module_feature->get_supported_method(stop_extra_threads_id, 0, stop_extra_threads), method_lookup_result::implemented);

    auto host1 = domain1->register_module_side_host(module_feature->get_handle(), 0);

    handle_type pack[100];
    object_handle releases[100];
    uint64_t release_counts[100];
    void* release_pointers[100];
    uint64_t release_count = 0;

    {
        handle_ptr<outbound_invocation<inbound_local_method_data>, module_to_kernel_invocation_handle> handle(
            domain1->get_outbound_invocations());
        auto* const request = static_cast<inbound_local_method_data*>(handle->get_inbound_method_data());
        request->set_request(host1, module_feature->get_handle(), stop_extra_threads_id, pack, releases, release_counts,
                             reinterpret_cast<char*>(release_pointers), &release_count);

        ASSERT_TRUE(domain1->start_invocation(handle.get_handle()));
        ASSERT_TRUE(handle);
        EXPECT_CALL(*domain1, send_request(_))
            .Times(1)
            .WillOnce(Return(domain_base::request_sending_status::pending_response));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());
        auto* local_data_passed = dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data());
        ASSERT_NE(local_data_passed, nullptr);

        object_handle target_object;
        interface_handle target_interface;
        method_id target_method;
        dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data())
            ->get_request_method_info(target_object, target_interface, target_method);
        ASSERT_EQ(target_object, host1);
        ASSERT_EQ(target_interface, module_feature->get_handle());
        ASSERT_EQ(target_method, method_id(stop_extra_threads_id));

        local_data_passed->write_exception_data(exception_handle::invalid);

        auto callee_side_handle = handle->get_callee_side_handle();
        ASSERT_TRUE(domain1->process_response(callee_side_handle));
        EXPECT_CALL(*domain1,
                    pass_response_data_to_module(domain1->get_outbound_invocations()->access(handle.get_handle())))
            .Times(1)
            .WillOnce(Return(response_sending_status::wait_for_modules_confirmation));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());

        exception_handle error_handle = static_cast<exception_handle>(static_cast<handle_type>(pack[0]));
        exception_kind error_kind = static_cast<exception_kind>(static_cast<handle_type>(pack[1]));
        ASSERT_EQ(error_handle, exception_handle::invalid);
        ASSERT_EQ(error_kind, static_cast<exception_kind>(static_cast<uint64_t>(invalid_interface_handle)));
    }
}

TEST(Invocation, ReturnValue)
{
    auto *const library = library::get_instance();

    kernel kernel(library);
    auto domain1 = make_shared<local_domain_mock>(&kernel);

    const auto *const module_feature = library->get_interface(interface_handle::system__module_feature);
    const method_base *get_extra_thread_count = nullptr;
    const method_id get_extra_thread_count_id = 0;
    ASSERT_EQ(module_feature->get_supported_method(get_extra_thread_count_id, 0, get_extra_thread_count), method_lookup_result::implemented);

    auto host1 = domain1->register_module_side_host(module_feature->get_handle(), 0);

    handle_type pack[100];
    object_handle releases[100];
    uint64_t release_counts[100];
    void* release_pointers[100];
    uint64_t release_count = 0;

    {
        pack[0] = 0;
        pack[1] = 0;
        pack[2] = 0;

        handle_ptr<outbound_invocation<inbound_local_method_data>, module_to_kernel_invocation_handle> handle(
            domain1->get_outbound_invocations());
        auto* const request = static_cast<inbound_local_method_data*>(handle->get_inbound_method_data());
        request->set_request(host1, module_feature->get_handle(), get_extra_thread_count_id, pack, releases, release_counts,
                             reinterpret_cast<char*>(release_pointers), &release_count);

        ASSERT_TRUE(domain1->start_invocation(handle.get_handle()));
        ASSERT_TRUE(handle);
        EXPECT_CALL(*domain1, send_request(_))
            .Times(1)
            .WillOnce(Return(domain_base::request_sending_status::pending_response));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());
        auto* local_data_passed = dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data());
        ASSERT_NE(local_data_passed, nullptr);

        object_handle target_object;
        interface_handle target_interface;
        method_id target_method;
        dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data())
            ->get_request_method_info(target_object, target_interface, target_method);
        ASSERT_EQ(target_object, host1);
        ASSERT_EQ(target_interface, module_feature->get_handle());
        ASSERT_EQ(target_method, method_id(get_extra_thread_count_id));

        local_data_passed->access_return_value().write(uint64_t(123));
        local_data_passed->write_exception_data(exception_handle::invalid);

        auto callee_side_handle = handle->get_callee_side_handle();
        ASSERT_TRUE(domain1->process_response(callee_side_handle));
        EXPECT_CALL(*domain1,
                    pass_response_data_to_module(domain1->get_outbound_invocations()->access(handle.get_handle())))
            .Times(1)
            .WillOnce(Return(response_sending_status::wait_for_modules_confirmation));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());

        exception_handle error_handle = static_cast<exception_handle>(static_cast<handle_type>(pack[0]));
        exception_kind error_kind = static_cast<exception_kind>(static_cast<handle_type>(pack[1]));
        uint64_t thread_count = pack[2];
        ASSERT_EQ(error_handle, exception_handle::invalid);
        ASSERT_EQ(error_kind, static_cast<exception_kind>(static_cast<uint64_t>(invalid_interface_handle)));
        ASSERT_EQ(thread_count, 123);
    }
}

TEST(Invocation, ThrowException)
{
    auto *const library = library::get_instance();

    kernel kernel(library);
    auto domain1 = make_shared<local_domain_mock>(&kernel);

    const auto *const module_feature = library->get_interface(interface_handle::system__module_feature);
    const method_base *get_extra_thread_count = nullptr;
    const method_id get_extra_thread_count_id = 0;
    ASSERT_EQ(module_feature->get_supported_method(get_extra_thread_count_id, 0, get_extra_thread_count), method_lookup_result::implemented);

    auto host1 = domain1->register_module_side_host(module_feature->get_handle(), 0);

    handle_type pack[100];
    object_handle releases[100];
    uint64_t release_counts[100];
    void* release_pointers[100];
    uint64_t release_count = 0;

    {
        pack[0] = 0;
        pack[1] = 0;
        pack[2] = 0;

        handle_ptr<outbound_invocation<inbound_local_method_data>, module_to_kernel_invocation_handle> handle(
            domain1->get_outbound_invocations());
        auto* const request = static_cast<inbound_local_method_data*>(handle->get_inbound_method_data());
        request->set_request(host1, module_feature->get_handle(), get_extra_thread_count_id, pack, releases, release_counts,
                             reinterpret_cast<char*>(release_pointers), &release_count);

        ASSERT_TRUE(domain1->start_invocation(handle.get_handle()));
        ASSERT_TRUE(handle);
        EXPECT_CALL(*domain1, send_request(_))
            .Times(1)
            .WillOnce(Return(domain_base::request_sending_status::pending_response));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());
        auto* local_data_passed = dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data());
        ASSERT_NE(local_data_passed, nullptr);

        object_handle target_object;
        interface_handle target_interface;
        method_id target_method;
        dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data())
            ->get_request_method_info(target_object, target_interface, target_method);
        ASSERT_EQ(target_object, host1);
        ASSERT_EQ(target_interface, module_feature->get_handle());
        ASSERT_EQ(target_method, method_id(get_extra_thread_count_id));

        local_data_passed->access_return_value().write(uint64_t(123));
        local_data_passed->write_exception_data(static_cast<exception_throw_mode>(0), static_cast<uint64_t>(exception_kind::system__exception));

        auto callee_side_handle = handle->get_callee_side_handle();
        ASSERT_TRUE(domain1->process_response(callee_side_handle));
        EXPECT_CALL(*domain1,
                    pass_response_data_to_module(domain1->get_outbound_invocations()->access(handle.get_handle())))
            .Times(1)
            .WillOnce(Return(response_sending_status::wait_for_modules_confirmation));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());

        exception_handle error_handle = static_cast<exception_handle>(static_cast<handle_type>(pack[0]));
        exception_kind error_kind = static_cast<exception_kind>(static_cast<handle_type>(pack[1]));
        uint64_t thread_count = pack[2];
        ASSERT_NE(error_handle, exception_handle::invalid);
        ASSERT_EQ(error_kind, exception_kind::system__exception);
        ASSERT_EQ(thread_count, 0);
    }
}

TEST(Invocation, Argument)
{
    auto *const library = library::get_instance();

    kernel kernel(library);
    auto domain1 = make_shared<local_domain_mock>(&kernel);

    const auto *const module_feature = library->get_interface(interface_handle::system__module_feature);
    const method_base *run_extra_thread = nullptr;
    const method_id run_extra_thread_id = 1;
    ASSERT_EQ(module_feature->get_supported_method(run_extra_thread_id, 0, run_extra_thread), method_lookup_result::implemented);

    auto host1 = domain1->register_module_side_host(module_feature->get_handle(), 0);

    handle_type pack[100];
    object_handle releases[100];
    uint64_t release_counts[100];
    void* release_pointers[100];
    uint64_t release_count = 0;

    {
        pack[0] = 0;
        pack[1] = 0;
        pack[2] = 123456;

        handle_ptr<outbound_invocation<inbound_local_method_data>, module_to_kernel_invocation_handle> handle(
            domain1->get_outbound_invocations());
        auto* const request = static_cast<inbound_local_method_data*>(handle->get_inbound_method_data());
        request->set_request(host1, module_feature->get_handle(), run_extra_thread_id, pack, releases, release_counts,
                             reinterpret_cast<char*>(release_pointers), &release_count);

        ASSERT_TRUE(domain1->start_invocation(handle.get_handle()));
        ASSERT_TRUE(handle);
        EXPECT_CALL(*domain1, send_request(_))
            .Times(1)
            .WillOnce(Return(domain_base::request_sending_status::pending_response));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());
        auto* local_data_passed = dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data());
        ASSERT_NE(local_data_passed, nullptr);

        object_handle target_object;
        interface_handle target_interface;
        method_id target_method;
        dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data())
            ->get_request_method_info(target_object, target_interface, target_method);
        ASSERT_EQ(target_object, host1);
        ASSERT_EQ(target_interface, module_feature->get_handle());
        ASSERT_EQ(target_method, method_id(run_extra_thread_id));
        
        uint64_t target_method_id = 0;
        local_data_passed->access_parameters(0).read(&target_method_id);
        EXPECT_EQ(target_method_id, 123456);

        local_data_passed->write_exception_data(exception_handle::invalid);

        auto callee_side_handle = handle->get_callee_side_handle();
        ASSERT_TRUE(domain1->process_response(callee_side_handle));
        EXPECT_CALL(*domain1,
                    pass_response_data_to_module(domain1->get_outbound_invocations()->access(handle.get_handle())))
            .Times(1)
            .WillOnce(Return(response_sending_status::wait_for_modules_confirmation));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());

        exception_handle error_handle = static_cast<exception_handle>(static_cast<handle_type>(pack[0]));
        exception_kind error_kind = static_cast<exception_kind>(static_cast<handle_type>(pack[1]));
        uint64_t thread_count = pack[2];
        ASSERT_EQ(error_handle, exception_handle::invalid);
        ASSERT_EQ(error_kind, static_cast<exception_kind>(static_cast<uint64_t>(invalid_interface_handle)));
        ASSERT_EQ(thread_count, 123456);
    }
}
