#include <sivkernel/interface.h>
#include <sivkernel/method.h>
#include <sivkernel/kernel.h>
#include <sivdk/exception_kind.h>
#include <sivkernel-mocks/domain_mock.h>
#include <sivdk/interface_handle.h>
#include <sivkernel-types/science/value_with_unit.h>
#include <gtest/gtest.h>
#include <gmock/gmock.h>

using namespace sivelkiria::kernel;
using namespace mocks;
using namespace sivelkiria::common;
using namespace sivelkiria::kernel::types::science;
using namespace data_pack;
using namespace std;
using namespace testing;


TEST(Invocation, ZeroFillSingleArgumentPads)
{
    kernel_mock kernel;
    auto domain1 = make_shared<local_domain_mock>(&kernel);

    method<void, value_with_unit<uint8_t, uint16_t, uint32_t>> method1(nullptr, nullptr, nullptr, nullptr, nullptr);

    interface iface1(static_cast<interface_handle>(0), {&method1}, {1}, {}, {});
    domain1->get_library()->add_interface(&iface1);

    auto host1 = domain1->register_module_side_host(iface1.get_handle(), 0);

    handle_type pack[100];
    object_handle releases[100];
    uint64_t release_counts[100];
    void* release_pointers[100];
    uint64_t release_count = 0;

    {
        std::fill_n(pack, sizeof(pack) / sizeof(pack[0]), 0xffffffffffffffff);

        *reinterpret_cast<uint8_t*>(pack + 2) = 0x01;
        *reinterpret_cast<uint16_t*>(pack + 3) = 0x0102;
        *reinterpret_cast<uint32_t*>(pack + 4) = 0x01020304;

        handle_ptr<outbound_invocation<inbound_local_method_data>, module_to_kernel_invocation_handle> handle(
            domain1->get_outbound_invocations());
        auto* const request = static_cast<inbound_local_method_data*>(handle->get_inbound_method_data());
        request->set_request(host1, iface1.get_handle(), 0, pack, releases, release_counts,
                             reinterpret_cast<char*>(release_pointers), &release_count);

        ASSERT_TRUE(domain1->start_invocation(handle.get_handle()));
        ASSERT_TRUE(handle);
        EXPECT_CALL(*domain1, send_request(_))
            .Times(1)
            .WillOnce(Return(domain_base::request_sending_status::pending_response));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());
        auto* local_data_passed = dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data());
        ASSERT_NE(local_data_passed, nullptr);

        object_handle target_object;
        interface_handle target_interface;
        method_id target_method;
        dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data())
            ->get_request_method_info(target_object, target_interface, target_method);
        ASSERT_EQ(target_object, host1);
        ASSERT_EQ(target_interface, iface1.get_handle());
        ASSERT_EQ(target_method, 0);

        const auto* raw_parameters = local_data_passed->access_parameters(0).get_data();

        raw_parameters += 1;
        for (auto byte = 0; byte < 7; ++byte)
        {
            EXPECT_EQ(*raw_parameters++, 0);
        }
        raw_parameters += 2;
        for (auto byte = 0; byte < 6; ++byte)
        {
            EXPECT_EQ(*raw_parameters++, 0);
        }
        raw_parameters += 4;
        for (auto byte = 0; byte < 4; ++byte)
        {
            EXPECT_EQ(*raw_parameters++, 0);
        }

        local_data_passed->write_exception_data(exception_handle::invalid);

        auto callee_side_handle = handle->get_callee_side_handle();
        ASSERT_TRUE(domain1->process_response(callee_side_handle));
        EXPECT_CALL(*domain1,
                    pass_response_data_to_module(domain1->get_outbound_invocations()->access(handle.get_handle())))
            .Times(1)
            .WillOnce(Return(response_sending_status::wait_for_modules_confirmation));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());

        exception_handle error_handle = static_cast<exception_handle>(static_cast<handle_type>(pack[0]));
        exception_kind error_kind = static_cast<exception_kind>(static_cast<handle_type>(pack[1]));
        ASSERT_EQ(error_handle, exception_handle::invalid);
        ASSERT_EQ(error_kind, static_cast<exception_kind>(static_cast<uint64_t>(invalid_interface_handle)));
    }
}


TEST(Invocation, ZeroFillVectorArgumentPads)
{
    kernel_mock kernel;
    auto domain1 = make_shared<local_domain_mock>(&kernel);

    method<void, std::vector<value_with_unit<uint8_t, uint16_t, uint32_t>>> method1(nullptr, nullptr, nullptr, nullptr, nullptr);

    interface iface1(static_cast<interface_handle>(0), {&method1}, {1}, {}, {});
    domain1->get_library()->add_interface(&iface1);

    auto host1 = domain1->register_module_side_host(iface1.get_handle(), 0);

    handle_type pack[100], elements[100];
    object_handle releases[100];
    uint64_t release_counts[100];
    void* release_pointers[100];
    uint64_t release_count = 0;

    {
        std::fill_n(pack, sizeof(pack) / sizeof(pack[0]), 0xffffffffffffffff);
        std::fill_n(elements, sizeof(elements) / sizeof(elements[0]), 0xffffffffffffffff);

        *reinterpret_cast<uint64_t*>(pack + 2) = 1;
        *reinterpret_cast<void**>(pack + 3) = &elements[0];

        *reinterpret_cast<uint8_t*>(elements) = 0x01;
        *reinterpret_cast<uint16_t*>(elements + 1) = 0x0102;
        *reinterpret_cast<uint32_t*>(elements + 2) = 0x01020304;

        handle_ptr<outbound_invocation<inbound_local_method_data>, module_to_kernel_invocation_handle> handle(
            domain1->get_outbound_invocations());
        auto* const request = static_cast<inbound_local_method_data*>(handle->get_inbound_method_data());
        request->set_request(host1, iface1.get_handle(), 0, pack, releases, release_counts,
                             reinterpret_cast<char*>(release_pointers), &release_count);

        ASSERT_TRUE(domain1->start_invocation(handle.get_handle()));
        ASSERT_TRUE(handle);
        EXPECT_CALL(*domain1, send_request(_))
            .Times(1)
            .WillOnce(Return(domain_base::request_sending_status::pending_response));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());
        auto* local_data_passed = dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data());
        ASSERT_NE(local_data_passed, nullptr);

        object_handle target_object;
        interface_handle target_interface;
        method_id target_method;
        dynamic_cast<outbound_local_method_data*>(
            domain1->get_outbound_invocations()->access(handle.get_handle())->get_outbound_method_data())
            ->get_request_method_info(target_object, target_interface, target_method);
        ASSERT_EQ(target_object, host1);
        ASSERT_EQ(target_interface, iface1.get_handle());
        ASSERT_EQ(target_method, 0);

        uint64_t array_size = 0;
        module_pointer array_pointer = module_pointer::null;
        local_data_passed->access_parameters(0).read(&array_size).read(&array_pointer);
        ASSERT_EQ(array_size, 1);

        const auto* raw_parameters = local_data_passed->sivelkiria::kernel::data_pack::data_pack::read_module_side_data<uint64_t>(array_pointer, 3).get_data();

        raw_parameters += 1;
        for (auto byte = 0; byte < 7; ++byte)
        {
            EXPECT_EQ(*raw_parameters++, 0);
        }
        raw_parameters += 2;
        for (auto byte = 0; byte < 6; ++byte)
        {
            EXPECT_EQ(*raw_parameters++, 0);
        }
        raw_parameters += 4;
        for (auto byte = 0; byte < 4; ++byte)
        {
            EXPECT_EQ(*raw_parameters++, 0);
        }

        local_data_passed->write_exception_data(exception_handle::invalid);

        auto callee_side_handle = handle->get_callee_side_handle();
        ASSERT_TRUE(domain1->process_response(callee_side_handle));
        EXPECT_CALL(*domain1,
                    pass_response_data_to_module(domain1->get_outbound_invocations()->access(handle.get_handle())))
            .Times(1)
            .WillOnce(Return(response_sending_status::wait_for_modules_confirmation));
        ASSERT_TRUE(domain1->get_scheduler()->process_single_action());

        exception_handle error_handle = static_cast<exception_handle>(static_cast<handle_type>(pack[0]));
        exception_kind error_kind = static_cast<exception_kind>(static_cast<handle_type>(pack[1]));
        ASSERT_EQ(error_handle, exception_handle::invalid);
        ASSERT_EQ(error_kind, static_cast<exception_kind>(static_cast<uint64_t>(invalid_interface_handle)));
    }
}

