#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <uchar.h>

#if defined(_WIN32)
#define MODULE_API_EXPORTS __declspec(dllexport)
#elif defined(__linux__)
#define MODULE_API_EXPORTS __attribute__((visibility("default")))
#else
#error "Unsupported platform"
#endif

typedef char bool;
typedef uint64_t handle_type;

const handle_type invalid_handle = UINT64_MAX;

const handle_type interface_module_execution = 0;
const handle_type interface_text_censorer = 18;
const handle_type interface_module_feature = 10;
const handle_type interface_writer = 16;

typedef enum
{
    invocation_status_success,
    invocation_status_exception,
    invocation_status_in_progress,
    invocation_status_invalid_handle,
    invocation_status_memory_error,
    invocation_status_waiting_not_allowed,
    invocation_status_already_being_checked
} invocation_status;

handle_type sivdk_register_host_object(handle_type interface, uint32_t version, void* extra_pointer);
bool sivdk_release_object_handle(handle_type handle);

handle_type sivdk_invoke(handle_type object, handle_type interface, uint32_t method, void* data, handle_type* released,
                         uint64_t* reference_counts, void** released_pointers, size_t* released_count);
invocation_status sivdk_invocation_wait(handle_type invocation);

struct interface_definition;
struct host;

struct interface_definition
{
    handle_type m_handle;
    struct interface_definition* m_parent;
    void (*m_invoke)(void* object_data, uint32_t method, void* data);
    void (*m_cleanup)(void* object_data);
};

struct interface_definition text_writer_definition;
struct interface_definition module_feature_definition;
struct interface_definition censorer_definition;
struct interface_definition module_execution_definition;

struct host
{
    handle_type m_handle;
    const struct interface_definition* m_interface;
    void* m_data;
    size_t m_references;
};

struct censorer
{
    handle_type m_output;
};

struct host* create_host(handle_type interface_handle, uint32_t version, struct interface_definition* interface,
                         void* data)
{
    struct host* const allocated = malloc(sizeof(struct host));

    allocated->m_handle = sivdk_register_host_object(interface_handle, version, allocated);
    allocated->m_interface = interface;
    allocated->m_data = data;
    allocated->m_references = 0;

    return allocated;
}

bool MODULE_API_EXPORTS release_host_references(handle_type handle, void* object_pointer, uint64_t references_to_release)
{
    struct host* const h = (struct host*)object_pointer;
    if (--h->m_references == 0)
    {
        (*h->m_interface->m_cleanup)(h->m_data);
        free(h);
    }
    return 1;
}

bool MODULE_API_EXPORTS handle_call(handle_type object_handle, void* object_pointer, handle_type interface,
                                    uint32_t method, void* data, handle_type invocation, void** invocation_pointer)
{
    struct
    {
        handle_type m_exception_instance;
        handle_type m_exception_type;
        handle_type m_next_field;
    }* d = data;
    d->m_exception_instance = invalid_handle;
    d->m_exception_type = invalid_handle;

    struct host* h = object_pointer;
    const struct interface_definition* i = h->m_interface;
    while (i && i->m_handle != interface)
        i = i->m_parent;
    assert(i);

    (*i->m_invoke)(h->m_data, method, &d->m_next_field);
    return 1;
}

void MODULE_API_EXPORTS release_inbound_call_data(handle_type handle, void* invocation_pointer)
{}

void invoke_text_writer(void* object_data, uint32_t method, void* data)
{
    switch (method)
    {
    case 0:
    {
        // write
        struct
        {
            uint64_t m_length;
            const char16_t* m_data;
        }* d = data;
        struct censorer* c = object_data;

        int found = 0;
        for (size_t i = 0; i + 4 <= d->m_length; ++i)
            if ((d->m_data[i] == u'T' || d->m_data[i] == u't') && d->m_data[i + 1] == u'e' && d->m_data[i + 2] == 's' &&
                d->m_data[i + 3] == 't')
            {
                char16_t* const changed = malloc(sizeof(char16_t) * d->m_length * 3);
                memcpy(changed, d->m_data, sizeof(char16_t) * i);
                changed[i] = d->m_data[i] == u'T' ? 'P' : 'p';
                memcpy(changed + i + 1, u"roduction", sizeof(char16_t) * 9);

                size_t last_in = i + 4;
                size_t last_out = i + 10;

                for (i = last_in; i + 4 <= d->m_length; ++i)
                {
                    if ((d->m_data[i] == u'T' || d->m_data[i] == u't') && d->m_data[i + 1] == u'e' &&
                        d->m_data[i + 2] == 's' && d->m_data[i + 3] == 't')
                    {
                        memcpy(changed + last_out, d->m_data + last_in, sizeof(char16_t) * (i - last_in));
                        last_out += (i - last_in);
                        changed[last_out] = d->m_data[i] == u'T' ? 'P' : 'p';
                        memcpy(changed + last_out + 1, u"roduction", sizeof(char16_t) * 9);
                        last_out += 10;
                        last_in = i + 4;
                        i += 3;
                    }
                }
                memcpy(changed + last_out, d->m_data + last_in, sizeof(char16_t) * (d->m_length - last_in));
                size_t end = last_out + (d->m_length - last_in);

                struct
                {
                    handle_type m_exception_instance;
                    handle_type m_exception_type;
                    uint64_t m_length;
                    const char16_t* m_data;
                } outgoing_invocation_data;
                outgoing_invocation_data.m_length = end;
                outgoing_invocation_data.m_data = changed;

                size_t release_count = 0;
                const handle_type outgoing_invocation_handle = sivdk_invoke(
                    c->m_output, interface_writer, 0, &outgoing_invocation_data, NULL, NULL, NULL, &release_count);
                const invocation_status result = sivdk_invocation_wait(outgoing_invocation_handle);
                if (result == invocation_status_exception)
                    sivdk_release_object_handle(outgoing_invocation_data.m_exception_instance);
                else
                    assert(result == invocation_status_success);

                free(changed);
                return;
            }

        struct
        {
            handle_type m_exception_instance;
            handle_type m_exception_type;
            uint64_t m_length;
            const char16_t* m_data;
        } outgoing_invocation_data;
        outgoing_invocation_data.m_length = d->m_length;
        outgoing_invocation_data.m_data = d->m_data;

        size_t release_count = 0;
        const handle_type outgoing_invocation_handle =
            sivdk_invoke(c->m_output, interface_writer, 0, &outgoing_invocation_data, NULL, NULL, NULL, &release_count);
        const invocation_status result = sivdk_invocation_wait(outgoing_invocation_handle);
        if (result == invocation_status_exception)
            sivdk_release_object_handle(outgoing_invocation_data.m_exception_instance);
        break;
    }
    default:
        assert(0);
    }
}

void cleanup_text_writer(void* object_data)
{
    struct censorer* const c = object_data;
    sivdk_release_object_handle(c->m_output);
}

void invoke_text_censorer(void* object_data, uint32_t method, void* data)
{
    switch (method)
    {
    case 0:
    {
        // get_censorer
        struct
        {
            handle_type m_created_writer;
            handle_type m_forward_to;
        }* d = data;

        struct censorer* c = malloc(sizeof(struct censorer));
        c->m_output = d->m_forward_to;
        struct host* h = create_host(interface_writer, 0, &text_writer_definition, c);
        d->m_created_writer = h->m_handle;
        break;
    }
    default:
        assert(0);
    }
}

void invoke_module_feature(void* object_data, uint32_t method, void* data)
{
    switch (method)
    {
    case 0:
    {
        // get_extra_thread_count
        struct
        {
            uint64_t m_count;
        }* d = data;
        d->m_count = 0;
        break;
    }
    case 1:
        // run_extra_thread
        break;
    case 2:
        // stop_extra_threads
        break;
    default:
        assert(0);
    }
}

void invoke_module_execution(void* object_data, uint32_t method, void* data)
{
    switch (method)
    {
    case 0:
    {
        // get_installation
        struct
        {
            handle_type m_handle;
        }* d = data;
        d->m_handle = invalid_handle;
        break;
    }
    case 1:
        // terminate
        break;
    case 2:
    {
        // get_types_sdk_version
        struct
        {
            uint64_t m_version;
        }* d = data;
        d->m_version = 0;
        break;
    }
    case 3:
    {
        // get_supported_features
        struct
        {
            uint64_t m_count;
            const handle_type* m_features;
        }* d = data;
        d->m_count = 1;
        d->m_features = &interface_text_censorer;
        break;
    }
    case 4:
    {
        // get_feature
        struct
        {
            handle_type m_instance;
            handle_type m_interface;
            handle_type m_requested_type;
        }* d = data;
        assert(d->m_requested_type == interface_text_censorer);
        d->m_interface = interface_text_censorer;

        struct host* h = create_host(interface_text_censorer, 0, &censorer_definition, NULL);
        d->m_instance = h->m_handle;
        break;
    }
    default:
        assert(0);
    }
}

void cleanup_stub(void* object_data)
{
    assert(object_data == NULL);
}

handle_type MODULE_API_EXPORTS initialize_module()
{
    text_writer_definition.m_handle = interface_writer;
    text_writer_definition.m_parent = NULL;
    text_writer_definition.m_invoke = &invoke_text_writer;
    text_writer_definition.m_cleanup = &cleanup_text_writer;

    module_feature_definition.m_handle = interface_module_feature;
    module_feature_definition.m_parent = NULL;
    module_feature_definition.m_invoke = &invoke_module_feature;
    module_feature_definition.m_cleanup = &cleanup_stub;

    censorer_definition.m_handle = interface_text_censorer;
    censorer_definition.m_parent = &module_feature_definition;
    censorer_definition.m_invoke = &invoke_text_censorer;
    censorer_definition.m_cleanup = &cleanup_stub;

    module_execution_definition.m_handle = interface_module_execution;
    module_execution_definition.m_parent = NULL;
    module_execution_definition.m_invoke = &invoke_module_execution;
    module_execution_definition.m_cleanup = &cleanup_stub;

    struct host* h = create_host(interface_module_execution, 0, &module_execution_definition, NULL);
    return h->m_handle;
}

void MODULE_API_EXPORTS handle_invocation_completion(handle_type invocation, void* pointer)
{}

void MODULE_API_EXPORTS returned_invalid_data_pack(handle_type invocation, void* invocation_pointer,
                                                       handle_type object, void* object_pointer,
                                                       handle_type interface, uint32_t method,
                                                       handle_type exception_instance,
                                                       handle_type exception_type)
{}

void MODULE_API_EXPORTS handle_event(handle_type subscription, void* pointer, const void* data)
{}
