﻿#include "entry_point.h"
#include <sivtypes/system/kernel.h>
#include <sivtypes/text/console.h>
#include <sivtypes/text/censorer.h>
#include <sivtypes/text/writer.h>
#include <sivtypes/messengers/filter.h>
#include <sivtypes/messengers/message.h>
#include <sivtypes/messengers/receiver.h>
#include <sivtypes/messengers/writer.h>
#include <sivtypes/messengers/message_source_proxy.h>
#include <sivtypes/messengers/message_source_proxy_provider.h>
#include <sivtypes/system/dynamic_library_domain.h>
#include <sivtypes/system/invalid_value_type_parameter.h>
#include <sivtypes/system/failed_loading_dynamic_library_module.h>
#include <sivtypes/system/dynamic_library_module_initialization_failed.h>
#include <sivdk/module_feature.h>
#include <sivdk/outgoing_call.h>
#include <iostream>
#include <cstdlib>
#include <atomic>
#include <thread>
#include <chrono>

using namespace std;
using namespace sivdk;
using namespace sivtypes::system;
using namespace sivtypes::messengers;
using namespace sivtypes::text;

static volatile atomic<size_type> messages_created = 0;
static volatile atomic<size_type> messages_destroyed = 0;
static volatile atomic<size_type> message_text_requested = 0;
static volatile atomic<size_type> message_sender_requested = 0;

entry_point_implementation::entry_point_implementation()
    : object(interface_handle::system__entry_point, get_interface_version())
{}

void entry_point_implementation::run(const shared_ptr<kernel>& kernel)
{
    try_loading_nonexisting_modules(kernel);
    run_message_conveyer(kernel);
}

void entry_point_implementation::try_loading_nonexisting_modules(const shared_ptr<kernel>& kernel)
{
    try
    {
        kernel->load_dynamic_library_module(u"NotAModule.dll");
        assert(false);
    }
    catch ([[maybe_unused]] const failed_loading_dynamic_library_module& error)
    {
        assert(error.get_filename() == u"NotAModule.dll");
    }
    try
    {
        kernel->load_dynamic_library_module(SIVELKIRIA_empty_module_PATH);
        assert(false);
    }
    catch ([[maybe_unused]] const dynamic_library_module_initialization_failed& error)
    {
        assert(error.get_filename() == SIVELKIRIA_empty_module_PATH);
    }
}

void entry_point_implementation::run_message_conveyer(const shared_ptr<kernel>& kernel)
{
    class message : public sivtypes::messengers::message
    {
    public:
        message(const u16string& sender_name, const u16string& text)
            : object(interface_handle::messengers__message, get_interface_version())
            , m_sender_name(sender_name)
            , m_text(text)
        {
            ++messages_created;
        }
        virtual ~message()
        {
            ++messages_destroyed;
        }

        u16string get_text() override
        {
            ++message_text_requested;
            return m_text;
        }
        u16string get_sender_name() override
        {
            ++message_sender_requested;
            return m_sender_name;
        }

    private:
        const u16string m_sender_name;
        const u16string m_text;
    };

    try
    {
        const auto message_chain_console = kernel->load_dynamic_library_module(SIVELKIRIA_message_chain_console_PATH);
        const auto console_writer_feature = message_chain_console->run_feature(module_feature_type::text__console);
        const auto console_writer_feature_typed = dynamic_pointer_cast<console>(console_writer_feature);
        const auto console_writer = console_writer_feature_typed->get_writer();

        const auto message_chain_filter_censorer =
            kernel->load_dynamic_library_module(SIVELKIRIA_message_chain_filter_censorer_PATH);
        const auto censorer_feature = message_chain_filter_censorer->run_feature(module_feature_type::text__censorer);
        const auto censorer_feature_typed = dynamic_pointer_cast<censorer>(censorer_feature);
        const auto censorer_writer = censorer_feature_typed->get_censorer(console_writer);

        const auto message_chain_filter_censorer_2 =
            kernel->load_dynamic_library_module(SIVELKIRIA_message_chain_c_censorer_PATH);
        const auto censorer_feature_2 = message_chain_filter_censorer_2->run_feature(module_feature_type::text__censorer);
        const auto censorer_feature_typed_2 = dynamic_pointer_cast<censorer>(censorer_feature_2);
        const auto censorer_writer_2 = censorer_feature_typed_2->get_censorer(censorer_writer);

        const auto messsage_to_text_converter =
            kernel->load_dynamic_library_module(SIVELKIRIA_message_chain_writer_PATH);
        const auto converter_feature = messsage_to_text_converter->run_feature(module_feature_type::messengers__writer);
        const auto converter_feature_typed = dynamic_pointer_cast<sivtypes::messengers::writer>(converter_feature);
        const auto converter_writer = converter_feature_typed->start_writing(censorer_writer_2);

        const auto filter1_feature = message_chain_filter_censorer->run_feature(module_feature_type::messengers__filter);
        const auto filter1_feature_typed = dynamic_pointer_cast<filter>(filter1_feature);
        const auto filter1 = filter1_feature_typed->start_filtering(converter_writer);

        const auto filter2_module = kernel->load_dynamic_library_module(SIVELKIRIA_message_chain_filter_2_PATH);
        const auto filter2_feature = filter2_module->run_feature(module_feature_type::messengers__filter);
        const auto filter2_feature_typed = dynamic_pointer_cast<filter>(filter2_feature);
        const auto filter2 = filter2_feature_typed->start_filtering(filter1);

        const auto proxy_provider_module = kernel->load_dynamic_library_module(SIVELKIRIA_message_chain_proxy_provider_PATH);
        const auto proxy_provider_feature = proxy_provider_module->run_feature(module_feature_type::messengers__message_source_proxy_provider);
        const auto proxy_provider_feature_typed = dynamic_pointer_cast<message_source_proxy_provider>(proxy_provider_feature);
        const auto proxy = proxy_provider_feature_typed->create_proxy();
        proxy->subscribe__new_message([&](auto*, const auto& message) {
            filter2->receive(message);
        });

        try
        {
            message_chain_console->run_feature(static_cast<module_feature_type>(static_cast<handle_type>(interface_handle::messengers__message)));
            assert(false);
        }
        catch (const invalid_value_type_parameter&)
        {
        }

        {
            SIVDK_CALL(message_source_proxy::push)
            message_1(proxy, make_shared<message>(u"Test", u"This is a test message."));
            SIVDK_CALL(message_source_proxy::push) message_2(proxy, make_shared<message>(u"Spammer", u"Some spam."));
            SIVDK_CALL(message_source_proxy::push)
            message_3(proxy, make_shared<message>(u"Your friend", u"I hate you, Mr. Bossman!"));
            SIVDK_CALL(message_source_proxy::push)
            message_4(proxy,
                      make_shared<message>(u"Иван Иванович Иванов", u"Это кириллический текст на русском языке."));

            message_1.start();
            message_2.start();
            message_3.start();
            message_4.start();

            message_1.wait();
            message_2.wait();
            message_3.wait();
            message_4.wait();
        }

        std::this_thread::sleep_for(std::chrono::seconds(1));

        assert(messages_created.load() == 4);
        assert(messages_destroyed.load() == 0);
        assert(message_text_requested.load() >= 4);
        assert(message_sender_requested.load() >= 4);
    }
    catch (std::exception& e)
    {
        cerr << "Unexpected exception: " << e.what() << endl;
        exit(2);
    }
    catch (...)
    {
        cerr << "Unexpected exception" << endl;
        exit(3);
    }
}

uint64_t entry_point_implementation::get_extra_thread_count()
{
    return 0;
}

void entry_point_implementation::run_extra_thread(uint64_t)
{}

void entry_point_implementation::stop_extra_threads()
{}
