#include <sivdk/outgoing_call.h>
#include <sivdk/module_functions.h>
#include <sivdk/call_data/proxy.h>
#include <gtest/gtest.h>
#include <sivdk-mocks/kernel_mock.h>
#include <sivdk-mocks/method_defines.h>

#include <sivtypes/system/module_feature.h>
#include <sivtypes/system/handle_call_failed.h>
#include <sivmocks/system/module_feature.h>
#include <sivdk/exception_kind.h>

using namespace sivdk;
using namespace call_data;
using namespace std;
using namespace mocks;
using namespace testing;
using namespace sivtypes::system;


TEST(SimpleCall, OutboundSimpleCall)
{
    object_proxy_handle_holder proxy_handle(object_handle(1));
    auto a = std::make_shared<module_feature__proxy>(&proxy_handle);

    {
        kernel_mock kernel;
        EXPECT_CALL(kernel, invoke_synchronously(object_handle(1), interface_handle::system__module_feature, GET_METHOD_ID(sivtypes::system::module_feature::stop_extra_threads), _, _, _, _, _, _))
            .Times(1)
            .WillOnce([](object_handle, interface_handle, method_id,
                         void* data, object_handle* released, uint64_t* reference_counts,
                         void** released_pointers, uint64_t* released_count,
                         module_to_kernel_invocation_handle* handle)
                {
                    EXPECT_NE(data, nullptr);
                    EXPECT_EQ(released, nullptr);
                    EXPECT_EQ(reference_counts, nullptr);
                    EXPECT_EQ(released_pointers, nullptr);
                    EXPECT_NE(released_count, nullptr);
                    EXPECT_EQ(*released_count, 0);
                    EXPECT_NE(handle, nullptr);

                    *handle = module_to_kernel_invocation_handle(123);
                    pack_field_iterator(data).write(exception_handle::invalid).write(exception_kind::invalid);

                    return invocation_status::success;
                }
            );
        a->stop_extra_threads();
    }

    {
        kernel_mock kernel;
        EXPECT_CALL(kernel, release_object_handle(object_handle(1))).Times(1).WillOnce(Return(true));
        a = nullptr;
    }
}

TEST(SimpleCall, OutboundReturnValue)
{
    object_proxy_handle_holder proxy_handle(object_handle(5));
    auto a = std::make_shared<module_feature__proxy>(&proxy_handle);

    {
        kernel_mock kernel;
        EXPECT_CALL(kernel, invoke_synchronously(object_handle(5), interface_handle::system__module_feature, GET_METHOD_ID(sivtypes::system::module_feature::get_extra_thread_count), _, _, _, _, _, _))
            .Times(1)
            .WillOnce([](object_handle, interface_handle, method_id,
                         void* data, object_handle* released, uint64_t* reference_counts,
                         void** released_pointers, uint64_t* released_count,
                         module_to_kernel_invocation_handle* handle)
                {
                    EXPECT_NE(data, nullptr);
                    EXPECT_EQ(released, nullptr);
                    EXPECT_EQ(reference_counts, nullptr);
                    EXPECT_EQ(released_pointers, nullptr);
                    EXPECT_NE(released_count, nullptr);
                    EXPECT_EQ(*released_count, 0);
                    EXPECT_NE(handle, nullptr);

                    *handle = module_to_kernel_invocation_handle(456);
                    pack_field_iterator(data).write(exception_handle::invalid).write(exception_kind::invalid).write(uint64_t(555));

                    return invocation_status::success;
                }
            );
        EXPECT_EQ(a->get_extra_thread_count(), 555);
    }

    {
        kernel_mock kernel;
        EXPECT_CALL(kernel, release_object_handle(object_handle(5))).Times(1).WillOnce(Return(true));
        a = nullptr;
    }
}

TEST(SimpleCall, OutboundException)
{
    object_proxy_handle_holder proxy_handle(object_handle(500));
    auto a = std::make_shared<module_feature__proxy>(&proxy_handle);

    {
        kernel_mock kernel;
        EXPECT_CALL(kernel, invoke_synchronously(object_handle(500), interface_handle::system__module_feature, GET_METHOD_ID(sivtypes::system::module_feature::get_extra_thread_count), _, _, _, _, _, _))
            .Times(1)
            .WillOnce([](object_handle, interface_handle, method_id,
                         void* data, object_handle* released, uint64_t* reference_counts,
                         void** released_pointers, uint64_t* released_count,
                         module_to_kernel_invocation_handle* handle)
                {
                    EXPECT_NE(data, nullptr);
                    EXPECT_EQ(released, nullptr);
                    EXPECT_EQ(reference_counts, nullptr);
                    EXPECT_EQ(released_pointers, nullptr);
                    EXPECT_NE(released_count, nullptr);
                    EXPECT_EQ(*released_count, 0);
                    EXPECT_NE(handle, nullptr);

                    *handle = module_to_kernel_invocation_handle(789);
                    pack_field_iterator(data).write(exception_handle(123)).write(exception_kind::system__handle_call_failed);

                    return invocation_status::exception;
                }
            );

        try
        {
            a->get_extra_thread_count();
            FAIL() << "No exception thrown";
        }
        catch (const handle_call_failed& e)
        {
            EXPECT_EQ(e.get_handle(), exception_handle(123));
            EXPECT_EQ(e.get_kind(), exception_kind::system__handle_call_failed);
            EXPECT_CALL(kernel, release_exception_handle(exception_handle(123)))
                .WillOnce(Return(true));
        }
    }

    {
        kernel_mock kernel;
        EXPECT_CALL(kernel, release_object_handle(object_handle(500))).Times(1).WillOnce(Return(true));
        a = nullptr;
    }
}

TEST(SimpleCall, InboundSimpleCall)
{
    std::shared_ptr<sivmocks::system::module_feature> a;

    {
        kernel_mock kernel;
        EXPECT_CALL(kernel, register_host_object(interface_handle::system__module_feature, 0, _))
            .WillOnce(Return(object_handle(1234)));
        a = std::make_shared<sivmocks::system::module_feature>();
    }

    EXPECT_EQ(a->get_handle(), object_handle(1234));

    {
        struct
        {
            alignas(handle_type) exception_throw_mode m_mode;
            alignas(handle_type) handle_type m_handle;
        } data {
            exception_throw_mode::no_exception,
            invalid_handle
        };
        void* invocation_pointer;

        EXPECT_CALL(*a, stop_extra_threads());
        handle_call(object_handle(1234), static_cast<object*>(a.get()), interface_handle::system__module_feature,
                    GET_METHOD_ID(sivtypes::system::module_feature::stop_extra_threads), &data, kernel_to_module_invocation_handle(1),
                    &invocation_pointer);

        EXPECT_EQ(data.m_mode, exception_throw_mode::no_exception);
        EXPECT_EQ(data.m_handle, invalid_handle);
    }

    {
        kernel_mock kernel;
        EXPECT_CALL(kernel, release_object_handle(object_handle(1234))).Times(1).WillOnce(Return(true));
        a = nullptr;
    }
}

TEST(SimpleCall, InboundReturnValue)
{
    std::shared_ptr<sivmocks::system::module_feature> a;

    {
        kernel_mock kernel;
        EXPECT_CALL(kernel, register_host_object(interface_handle::system__module_feature, 0, _))
            .WillOnce(Return(object_handle(1234)));
        a = std::make_shared<sivmocks::system::module_feature>();
    }

    EXPECT_EQ(a->get_handle(), object_handle(1234));

    {
        struct
        {
            alignas(handle_type) exception_throw_mode m_mode;
            alignas(handle_type) handle_type m_handle;
            alignas(handle_type) uint64_t m_count;
        } data {
            exception_throw_mode::no_exception,
            invalid_handle,
            0
        };
        void* invocation_pointer;

        EXPECT_CALL(*a, get_extra_thread_count()).WillOnce(Return(531));
        handle_call(object_handle(1234), static_cast<object*>(a.get()), interface_handle::system__module_feature,
                    GET_METHOD_ID(sivtypes::system::module_feature::get_extra_thread_count), &data, kernel_to_module_invocation_handle(1),
                    &invocation_pointer);

        EXPECT_EQ(data.m_mode, exception_throw_mode::no_exception);
        EXPECT_EQ(data.m_handle, invalid_handle);
        EXPECT_EQ(data.m_count, 531);
    }

    {
        kernel_mock kernel;
        EXPECT_CALL(kernel, release_object_handle(object_handle(1234))).Times(1).WillOnce(Return(true));
        a = nullptr;
    }
}

TEST(SimpleCall, InboundException)
{
    std::shared_ptr<sivmocks::system::module_feature> a;

    {
        kernel_mock kernel;
        EXPECT_CALL(kernel, register_host_object(interface_handle::system__module_feature, 0, _))
            .WillOnce(Return(object_handle(1234)));
        a = std::make_shared<sivmocks::system::module_feature>();
    }

    EXPECT_EQ(a->get_handle(), object_handle(1234));

    {
        struct
        {
            alignas(handle_type) exception_throw_mode m_mode;
            alignas(handle_type) exception_kind m_handle;
            alignas(handle_type) uint64_t m_count;
        } data {
            exception_throw_mode::no_exception,
            exception_kind::invalid,
            0
        };
        void* invocation_pointer;

        kernel_mock kernel;

        EXPECT_CALL(*a, get_extra_thread_count()).WillOnce([]() -> uint64_t{ throw handle_call_failed(); });
        handle_call(object_handle(1234), static_cast<object*>(a.get()), interface_handle::system__module_feature,
                    GET_METHOD_ID(sivtypes::system::module_feature::get_extra_thread_count), &data, kernel_to_module_invocation_handle(1),
                    &invocation_pointer);

        EXPECT_EQ(data.m_mode, static_cast<exception_throw_mode>(0));
        EXPECT_EQ(data.m_handle, exception_kind::system__handle_call_failed);
        EXPECT_EQ(data.m_count, 0);
    }

    {
        kernel_mock kernel;
        EXPECT_CALL(kernel, release_object_handle(object_handle(1234))).Times(1).WillOnce(Return(true));
        a = nullptr;
    }
}

